package com.recordingblogs;

import java.io.*;
import javax.sound.sampled.*;

/**
 * An implementation of a wave file.  Note that this is probably not the best way 
 * a wave file should be implemented.  The wave file format allows chunks
 * within chunks.  This simplistic implementation, on the other hand, expects that
 * the format and data chunks will be right at the top of the wave file, under 
 * the RIFF chunk.  This is true of most wave files, although not all.  Since 
 * chunks can be nested, it may be best to implement a wave file with a stack
 */

public class WaveFile extends RandomAccessFile
{
	/**
	 * The position of the data chunk, after the chunk name 'data' and the
	 * four byte chunk size
	 */
    private long m_datapointer;
    
    /**
     * The format of this wave file
     */
    private AudioFormat m_format;

    /**
     * The one and only constructor
     * @param name The full name (with path) of the wave file
     * @param rw In this implementation, "r"
     * @throws IOException The constructor for RandomAccessFile may
     *                     throw an IOException
     */
	public WaveFile(String name, String rw) throws IOException
	{
		super(name, rw);
		m_datapointer = -1;
	}

	/**
	 * Get the audio format
	 * @return The audio format
	 */
	public AudioFormat getFormat()
	{
		return m_format;
	}

	/**
	 * Read strings (the names of chunks) from the wave file
	 * @param size The size of the string to be read
	 * @return The string that was read
	 * @throws IOException  Reading from a RandomAccessFile may throw an
	 *                      IOException
	 */
    public String readRIFFString(int size) throws IOException
    {
        String string = new String();
        for(int i = 0; i < size; i++)
        {
            char c = (char) readByte();
            string = string + c;
        }
        return string;
    }

    /**
     * Read an unsigned short, mostly used to read pieces of the format chunk
     * @return The short that was read
	 * @throws IOException  Reading from a RandomAccessFile may throw an
	 *                      IOException
     */
    public int readRIFFUnsignedShort() throws IOException
    {
        byte b0 = readByte();
        byte b1 = readByte();
        return b0 + (b1 << 4);
    }

    /**
     * Read an unsigned integer, used to read pieces of the format chunk and
     * the sizes of chunks
     * @return The integer that was read
	 * @throws IOException  Reading from a RandomAccessFile may throw an
	 *                      IOException
     */
    public long readRIFFUnsignedInt() throws IOException
    {
        long b0 = readUnsignedByte();
        long b1 = readUnsignedByte();
        long b2 = readUnsignedByte();
        long b3 = readUnsignedByte();
        return b0 + (b1 << 8) + (b2 << 16) + (b3 << 24);
    }

    /**
     * Open the wave file, check its form and format, and find the positions
     * of the format chunk data and the data chunk data
     * @return true, if there were no errors; false, otherwise
     */
    public boolean openRead()
    {
        try
        {
        	// start from the beginning
            seek(0);

            // check for a RIFF header
            String temp = readRIFFString(4);
            if(temp.compareTo("RIFF") != 0)					//$NON-NLS-1$
            {
                System.out.println("Error in WaveFile::openRead: Not a valid RIFF file");	//$NON-NLS-1$
                close();
                return false;
            }

            // read the size of the RIFF chunk
            long size = readRIFFUnsignedInt();
            
            // check the subtype of the RIFF chunk
            temp = readRIFFString(4);
            if(temp.compareTo("WAVE") != 0)	//$NON-NLS-1$
            {
            	System.out.println("Error in WaveFile::openRead: Not a valid WAVE file");	//$NON-NLS-1$
                close();
                return false;
            }
            
            m_datapointer = -1;
            long fmtpointer = -1;
            
            // Look for a fmt chunk or a data chunk.  If these chunks are
            // not present, this code will throw an exception caught below
            while(m_datapointer < 0 || fmtpointer < 0)
            {
	            temp = readRIFFString(4);
	            if(temp.compareTo("fmt ") == 0)	//$NON-NLS-1$
	            	fmtpointer = getFilePointer() + 4;
	            if(temp.compareTo("data") == 0)	//$NON-NLS-1$
	            	m_datapointer = getFilePointer() + 4;
	            size = readRIFFUnsignedInt();
	            seek(getFilePointer() + size);
            }

            // read the format chunk
            seek(fmtpointer);
            short formatType = (short) readRIFFUnsignedShort();
            short channels = (short) readRIFFUnsignedShort();
            long sampleRate = readRIFFUnsignedInt();
            readRIFFUnsignedInt();    // reads bytes per second
            readRIFFUnsignedShort();  // reads bytes per sample
            short bitsPerChannel = (short) readRIFFUnsignedShort();

            // check the format.  For this application, we want a PCM wave
            // with the sampling rate 44100 Hz and the 16-bit resolution
            if (formatType != 1 || sampleRate != 44100 || bitsPerChannel != 16)
            {
            	System.out.println("Error in WaveFile::openRead: Not the right wave file format");	//$NON-NLS-1$
                close();
                return false;
            }
            
            // alternatively, we can read the format chunk with the following
            // AudioSystem.getAudioFileFormat(new File(wavedlg.getFileName().getText()));

            // remember the format
            m_format = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, bitsPerChannel, channels, (bitsPerChannel / 8) * channels, sampleRate, false);
        }
        catch(IOException e)
        {
            System.out.println("Exception in MyWaveFile::openRead: " + e);	//$NON-NLS-1$
            return false;
        }
        
        return true;
    }

    /**
     * Prepare to start playing, by positioning the file pointer at the
     * beginning of the data chunk (after the name of the chunk and the size
     * of the chunk, at the actual audio data)
     * @return true, if there were no errors; false, otherwise
     */
    public boolean startPlay()
    {
    	try
    	{
    		seek(m_datapointer);
    	}
    	catch (IOException e)
    	{
            System.out.println("Exception in WaveFile::startPlay: " + e);	//$NON-NLS-1$
    		return false;
    	}
    	
    	return true;
    }

    /**
     * Get a buffer of audio data
     * @param buffer The buffer to be filled with audio data
     * @return true, if there were no errors; false, otherwise
     */
    public boolean getData(AudioBuffer buffer)
    {
    	try
    	{
    		int actualsize = this.read(buffer.getBuffer(), 0, buffer.getSize());
    		buffer.setActualSize(actualsize);
    	}
    	catch (IOException e)
    	{
            System.out.println("Exception in WaveFile::getData: " + e);	//$NON-NLS-1$
    		return false;
    	}
    	
    	return true;
    }
}
